import tensorflow.compat.v1 as tf1
import tensorflow as tf



class DeepCas9(object):

    def __init__(self, filter_size, filter_num, node_1=80, node_2=60, l_rate=0.005):
        # filter_size = [3, 5, 7]
        # filter_num = [100, 70, 40]
        length = 30
        self.inputs = tf1.placeholder(tf1.float32, [None, 1, length, 4])
        self.targets = tf1.placeholder(tf1.float32, [None, 1])
        self.is_training = tf1.placeholder(tf1.bool)

        def create_new_conv_layer(input_data, num_input_channels, num_filters, filter_shape, pool_shape, name):
            # setup the filter input shape for tf.nn.conv_2d
            conv_filt_shape = [filter_shape[0], filter_shape[1], num_input_channels,
                               num_filters]

            # initialise weights and bias for the filter
            weights = tf1.Variable(tf1.truncated_normal(conv_filt_shape, stddev=0.03),
                                   name=name + '_W')
            bias = tf1.Variable(tf1.truncated_normal([num_filters]), name=name + '_b')

            # setup the convolutional layer operation
            out_layer = tf.nn.conv2d(input_data, weights, [1, 1, 1, 1], padding='VALID')
            #out_layer = tf1.nn.conv2d(input_data, weights, [1, 1, 1, 1], padding='VALID')

            # add the bias
            out_layer += bias

            # apply a ReLU non-linear activation
            #out_layer = tf1.layers.dropout(tf1.nn.relu(out_layer), 0.3, self.is_training)
            out_layer = tf.keras.layers.Dropout(rate=0.3)(tf.nn.relu(out_layer))

            # now perform max pooling
            ksize = [1, pool_shape[0], pool_shape[1], 1]
            strides = [1, 1, 2, 1]
            out_layer = tf.nn.avg_pool(out_layer, ksize=ksize, strides=strides,
                                        padding='SAME')
            return out_layer

        # def end: create_new_conv_layer

        L_pool_0 = create_new_conv_layer(self.inputs, 4, filter_num[0], [1, filter_size[0]], [1, 2], name='conv1')
        L_pool_1 = create_new_conv_layer(self.inputs, 4, filter_num[1], [1, filter_size[1]], [1, 2], name='conv2')
        L_pool_2 = create_new_conv_layer(self.inputs, 4, filter_num[2], [1, filter_size[2]], [1, 2], name='conv3')

        with tf1.variable_scope('Fully_Connected_Layer1'):
            layer_node_0 = int((length - filter_size[0]) / 2) + 1
            node_num_0 = layer_node_0 * filter_num[0]
            layer_node_1 = int((length - filter_size[1]) / 2) + 1
            node_num_1 = layer_node_1 * filter_num[1]
            layer_node_2 = int((length - filter_size[2]) / 2) + 1
            node_num_2 = layer_node_2 * filter_num[2]

            L_flatten_0 = tf.reshape(L_pool_0, [-1, node_num_0])
            L_flatten_1 = tf.reshape(L_pool_1, [-1, node_num_1])
            L_flatten_2 = tf.reshape(L_pool_2, [-1, node_num_2])
            L_flatten = tf.concat([L_flatten_0, L_flatten_1, L_flatten_2], 1, name='concat')

            node_num = node_num_0 + node_num_1 + node_num_2
            W_fcl1 = tf1.get_variable("W_fcl1", shape=[node_num, node_1])
            B_fcl1 = tf1.get_variable("B_fcl1", shape=[node_1])
            L_fcl1_pre   = tf.nn.bias_add(tf.matmul(L_flatten, W_fcl1), B_fcl1)
            L_fcl1       = tf.nn.relu(L_fcl1_pre)
            L_fcl1_drop  = tf.keras.layers.Dropout(rate=0.3)(L_fcl1)


        with tf1.variable_scope('Fully_Connected_Layer2'):
            W_fcl2 = tf1.get_variable("W_fcl2", shape=[node_1, node_2])
            B_fcl2 = tf1.get_variable("B_fcl2", shape=[node_2])
            L_fcl2_pre = tf.nn.bias_add(tf.matmul(L_fcl1_drop, W_fcl2), B_fcl2)
            L_fcl2 = tf.nn.relu(L_fcl2_pre)
            L_fcl2_drop = tf.keras.layers.Dropout(rate=0.3)(L_fcl2)


        with tf1.variable_scope('Output_Layer'):
            W_out = tf1.get_variable("W_out",
                                     shape=[node_2, 1])  # , initializer=tf.contrib.layers.xavier_initializer())
            B_out = tf1.get_variable("B_out", shape=[1])  # , initializer=tf.contrib.layers.xavier_initializer())
            self.outputs = tf.nn.bias_add(tf.matmul(L_fcl2_drop, W_out), B_out)

        # Define loss function and optimizer
        self.obj_loss = tf.reduce_mean(tf.square(self.targets - self.outputs))
        self.optimizer = tf1.train.AdamOptimizer(l_rate).minimize(self.obj_loss)
    # def end: def __init__



# == Native Modules
import re
import sys
import sys
# == Installed Modules
from Bio import SeqIO
# == Project Modules


def main():
	INPUT_FASTA = sys.argv[1]
	OUTPUT_FASTA = sys.argv[2]

	# INPUT_CHROMOSOME = "/home/ubuntu/genomes/toy_genome.fna"

	record = SeqIO.parse(open(INPUT_FASTA), "fasta")
	changed_recs = []
	for rec in record:
		a = re.search(r"Homo sapiens chromosome (\w+), GRCh38 reference primary assembly", rec.description)
		try:
			rec.id = f"chr{a.group(1)}"
			rec.name = f"chr{a.group(1)}"
			rec.description = ""
		except AttributeError:
			a = re.search(r"Homo sapiens (m)itochondrion", rec.description)
			try:
				mit_id = a.group(1)
				rec.id = f"chrMT"
				rec.name = f"chrMT"
				rec.description = ""
			except AttributeError:
				continue
		changed_recs.append(rec)

	SeqIO.write(changed_recs, OUTPUT_FASTA, 'fasta')


if __name__ == "__main__":
	main()



# #== src/prog/ncbi_cross_database.py ==
queries = ["BCL11A"]
in_path = "/Users/bellieny/projects/mEdit/hgvs_input.csv"
entrez_login = 'thedoudnalab@gmail.com'
db_from = 'gene'
dbto = 'clinvar'
final_df_orient = 'columns'

# #==  ==

# #==  ==



# == Native Modules
import pickle
from collections import defaultdict
import shutil
# == Installed Modules
import pandas as pd
# == Project Modules


def parse_bulge(tmp_processing_filename):
	isreversed = False
	id_dict = {}
	with open(tmp_processing_filename, 'r') as f:
		chrom_path = f.readline()
		pattern, bulge_dna, bulge_rna = f.readline().strip().split()
		for i in range(int(len(pattern) / 2)):
			if pattern[i] == 'N' and pattern[len(pattern) - i - 1] != 'N':
				isreversed = False
				break
			elif pattern[i] != 'N' and pattern[len(pattern) - i - 1] == 'N':
				isreversed = True
				break
		bulge_dna, bulge_rna = int(bulge_dna), int(bulge_rna)
		targets = [line.strip().split() for line in f]
		rnabulge_dic = defaultdict(lambda: [])
		bg_tgts = defaultdict(lambda: set())
		for raw_target, mismatch, gid in targets:
			if isreversed:
				target = raw_target.lstrip('N')
				len_pam = len(raw_target) - len(target)
				bg_tgts['N' * len_pam + target + 'N' * bulge_dna].add(mismatch)
				id_dict['N' * len_pam + target + 'N' * bulge_dna] = gid
				for bulge_size in range(1, bulge_dna + 1):
					for i in range(1, len(target)):
						bg_tgt = 'N' * len_pam + target[:i] + 'N' * bulge_size + target[i:] + 'N' * (
									bulge_dna - bulge_size)
						bg_tgts[bg_tgt].add(mismatch)
						id_dict[bg_tgt] = gid

				for bulge_size in range(1, bulge_rna + 1):
					for i in range(1, len(target) - bulge_size):
						bg_tgt = 'N' * len_pam + target[:i] + target[i + bulge_size:] + 'N' * (bulge_dna + bulge_size)
						bg_tgts[bg_tgt].add(mismatch)
						rnabulge_dic[bg_tgt].append((i, int(mismatch), target[i:i + bulge_size]))
						id_dict[bg_tgt] = gid
			else:
				target = raw_target.rstrip('N')
				len_pam = len(raw_target) - len(target)
				bg_tgts['N' * bulge_dna + target + 'N' * len_pam].add(mismatch)
				id_dict['N' * bulge_dna + target + 'N' * len_pam] = gid
				for bulge_size in range(1, bulge_dna + 1):
					for i in range(1, len(target)):
						bg_tgt = 'N' * (bulge_dna - bulge_size) + target[:i] + 'N' * bulge_size + target[
																								  i:] + 'N' * len_pam
						bg_tgts[bg_tgt].add(mismatch)
						id_dict[bg_tgt] = gid

				for bulge_size in range(1, bulge_rna + 1):
					for i in range(1, len(target) - bulge_size):
						bg_tgt = 'N' * (bulge_dna + bulge_size) + target[:i] + target[i + bulge_size:] + 'N' * len_pam
						bg_tgts[bg_tgt].add(mismatch)
						rnabulge_dic[bg_tgt].append((i, int(mismatch), target[i:i + bulge_size]))
						id_dict[bg_tgt] = gid
		if isreversed:
			seq_pam = pattern[:len_pam]
		else:
			seq_pam = pattern[-len_pam:]
	return isreversed, chrom_path, seq_pam, rnabulge_dic, id_dict, len_pam, pattern, bulge_dna, bg_tgts


def write_casoff_input(tmp_processing_filename, casoff_in_path, bulge_check: bool):
	'''
	 The cas-offinder off-line package contains a bug that doesn't allow bulges
	This script is partially a wrapper for cas-offinder to fix this bug
	 created by...
	https://github.com/hyugel/cas-offinder-bulge
	'''
	if bulge_check:
		(isreversed,
		 chrom_path,
		 seq_pam,
		 rnabulge_dic,
		 id_dict,
		 len_pam,
		 pattern,
		 bulge_dna,
		 bg_tgts) = parse_bulge(tmp_processing_filename)

		with open(casoff_in_path, 'w') as f:
			f.write(chrom_path)
			if isreversed:
				f.write(pattern + bulge_dna*'N' + '\n')
			else:
				f.write(bulge_dna*'N' + pattern + '\n')
			cnt = 0
			for tgt, mismatch in bg_tgts.items():
				f.write(tgt + ' ' + str(max(mismatch)) + ' ' + '\n')
				cnt += 1
	if not bulge_check:
		shutil.copy2(tmp_processing_filename, casoff_in_path)


def check_bulge(casoff_params):
	if casoff_params[1:3] == (0, 0):
		bulge = False
	else:
		bulge = True
	return bulge


def parse_casoff_support(tmp_processing_filename, fasta_fname, pam, pamISfirst, guidelen, guides, gnames, casoff_params):
	## create input file for cas-offinder
	mm, RNAbb, DNAbb, PU = casoff_params

	with open(tmp_processing_filename, 'w') as f:
		f.writelines(fasta_fname + "\n")
		line = 'N' * guidelen

		if pamISfirst:
			line = f"{pam}{line} {DNAbb} {RNAbb}" + "\n"
		else:
			line = f"{line}{pam} {DNAbb} {RNAbb}" + "\n"
		f.writelines(line)
		dpam = 'N' * len(pam)
		for grna, gname in zip(guides, gnames):
			if pamISfirst:
				line = f"{dpam}{grna} {mm} {gname}" + "\n"
			else:
				line = f"{grna}{dpam} {mm} {gname}" + "\n"
			f.writelines(line)


def main():
	# SNAKEMAKE IMPORTS
	# === Inputs ===
	guides_report_per_editor_path = str(snakemake.input.guides_per_editor_path)
	guide_search_params = str(snakemake.input.guide_search_params)
	assembly_reference_path = str(snakemake.input.decompressed_assembly_symlink)
	# snv_site_info = str(snakemake.input.snv_site_info)
	# annote_path = str(snakemake.params.annote_path)
	# === Outputs ===
	casoff_support_path = str(snakemake.output.casoff_support)
	casoff_input_path = str(snakemake.output.casoff_input)
	# === Params ===
	rna_bulge = str(snakemake.params.rna_bulge)
	dna_bulge = str(snakemake.params.dna_bulge)
	maximum_mismatches = str(snakemake.params.max_mismatch)
	PU = str(snakemake.params.casoff_accelerator)
	# === Wildcards ===
	editing_tool = str(snakemake.wildcards.editing_tool)

	# === Guide search params ===
	search_params = pickle.load(open(guide_search_params, 'rb'))
	guides_report_per_editor = pickle.load(open(guides_report_per_editor_path, 'rb'))

	pam, pamISfirst, guidelen = search_params[editing_tool][0:3]
	guides, gnames = list(guides_report_per_editor.gRNA), list(guides_report_per_editor.Guide_ID)

	# Check bulge based on pre-defined Cas-Offinder params
	casoff_params = (maximum_mismatches, rna_bulge, dna_bulge, PU)
	bulge_check = check_bulge(casoff_params)

	parse_casoff_support(casoff_support_path,
						 assembly_reference_path,
						 pam,
						 pamISfirst,
						 guidelen,
						 guides,
						 gnames,
						 casoff_params)

	# cas_offinder_bulge FUNCTION STARTS HERE
	write_casoff_input(casoff_support_path, casoff_input_path, bulge_check)


if __name__ == "__main__":
	main()



# == Native Modules
from os.path import abspath
import pickle
import subprocess
from pathlib import Path
# == Installed Modules
import yaml
# == Project Modules
from prog.medit_lib import (
	export_guides_by_editor,
	file_exists,
	group_guide_table,
	launch_shell_cmd,
	project_file_path,
	set_export,
	write_yaml_to_file
)


def offtarget_prediction(args, jobtag):
	# == Load Run Parameters values ==
	user_jobtag = args.user_jobtag
	root_dir = abspath(args.output)
	db_path_full = f"{abspath(args.db_path)}/medit_database"
	editing_tool_request = args.select_editors
	# == Load SLURM-related values ==
	ncores = args.ncores
	maxtime = args.maxtime
	parallel_processes = args.parallel_processes
	dry_run = args.dry_run

	# == Set export paths tied to the SMK pipeline ==
	config_dir_path = f"{root_dir}/config"

	if editing_tool_request:
		editing_tool_request = list(editing_tool_request.split(","))

	# == Set import/export paths for dynamic YAML files ==
	config_db_path = f"{db_path_full}/config_db/config_db.yaml"
	dynamic_config_off_path = f"{config_dir_path}/config_off_{jobtag}.yaml"  # OffTarget config generated in this program
	dynamic_config_guidepred_path = f"{config_dir_path}/config_{jobtag}.yaml"  # Main config generated by guide_prediction.py

	# == Check existence of the config file from guide_prediciton ==
	if not file_exists(dynamic_config_guidepred_path):
		raise (f"This program depends on the configuration file associated with the JOBTAG {jobtag}. "
			   f"Such file wasn't found on this parent folder: {root_dir}. "
			   f"Make sure to provide the same path as the OUTPUT argument given to 'medit guide_prediction' ")

	#   == Load template configuration files ==
	config_template_path = project_file_path("smk.config", "medit_offtarget.yaml")
	config_cluster_path = project_file_path("smk.config", "medit_cluster.yaml")

	# == Define dynamic SMK call variables ==
	cluster_smk_setup = ''
	smk_verbosity = True
	smk_run_triggers = ''
	dryrun_setup = ''

	#   == Check the dry run request
	if dry_run:
		dryrun_setup = '-n'
	if user_jobtag:
		smk_run_triggers = '--rerun-triggers "mtime"'
	#   == Check the request to run on a cluster
	if parallel_processes:
		# --> Upon SLURM run request, the guide_prediction.smk is split in two separate runs
		# --> That's because samtool's conda package crashes on a libcrypto error when
		#       it's deployed by snakemake on a SLURM node
		cluster_smk_setup = ('--cluster "sbatch -t {cluster.time} -n {cluster.cores}" '
							 f'--cluster-config  {config_cluster_path}')

	# ->=== CONFIG FILES IMPORT ===<-
	with open(dynamic_config_guidepred_path, 'r') as dynamic_config_handle:
		dynamic_config_guidepred = yaml.safe_load(dynamic_config_handle)
	with open(config_template_path, 'r') as config_handle:
		config_template = yaml.safe_load(config_handle)

	# === Import Variables from Configuration File ===
	run_name = str(dynamic_config_guidepred['run_name'])
	mode = str(dynamic_config_guidepred['processing_mode'])
	root_dir = str(dynamic_config_guidepred['output_directory'])
	# == mEdit offtarget prediction will only process one reference genome in this version ==
	sequence_id = str(dynamic_config_guidepred['sequence_id'][0])

	# == Set output paths ==
	guides_per_editor_path = str(
		f"{root_dir}/{mode}/jobs/{run_name}/guide_prediction-{sequence_id}/offtarget_prediction")

	# == Recover Guide Prediction filepath ==
	guides_report_path = Path(f"{root_dir}/{mode}/jobs/{run_name}/"
							  f"guide_prediction-{sequence_id}/guides_report_ref/Guides_found.csv")

	# == Group guides by editor and export DF as pickles by editor
	grouped_guide_dict = group_guide_table(guides_report_path, editing_tool_request)
	editors_list = export_guides_by_editor(grouped_guide_dict, guides_per_editor_path)

	# === Export Variables to Configuration File ===
	config_template['guides_per_editor_path'] = guides_per_editor_path
	config_template['editors_list'] = editors_list
	config_template['tmp_processing_casoff'] = f"{root_dir}/{config_template['tmp_processing_casoff']}"

	# == Set the temporary directory up for CasOffinder ==
	set_export(config_template['tmp_casoff_path'])

	# === Write YAML configs to mEdit Root Directory ===
	write_yaml_to_file(config_template, dynamic_config_off_path)

	# === Invoke SMK Pipelines ===
	print("# == Calling Off-Target Prediction pipeline == #")
	try:
		# --> When cluster submission is switched on,
		launch_shell_cmd(f"snakemake "
						 f"--snakefile {project_file_path('smk.pipelines', 'offtarget_prediction.smk')} "
						 f"{smk_run_triggers} "
						 f"-j {ncores} "
						 f"{cluster_smk_setup} "
						 f"--configfile {config_db_path} "
						 f"{dynamic_config_guidepred_path} {dynamic_config_off_path} "
						 f"--use-conda "
						 f"--rerun-incomplete "
						 f"{dryrun_setup} ",
						 smk_verbosity
						 )
	except subprocess.CalledProcessError as e:
		print(f"Error: {e}")
	except ValueError:
		print(f"Process completed in a previous run. Moving to the next one...")



# == Native Modules
import pickle
from os.path import abspath, isdir
# == Installed Modules
import yaml
# == Project Modules
from prog.medit_lib import file_exists


def ls_editors(args):
	# == Load Run Parameters values ==
	# print_editors = args.editors
	# print_base_editors = args.base_editors

	# === Load Database Path ===
	db_path_full = f"{abspath(args.db_path)}/medit_database"
	config_db_path = f"{db_path_full}/config_db/config_db.yaml"

	if not file_exists(db_path_full):
		print("The database path directory could not be found.")
		exit(0)

	# === Load configuration file ===
	with open(config_db_path, 'r') as config_handle:
		config_db_obj = yaml.safe_load(config_handle)

	# === Load Editors Lists From Path Specified on config_db.yaml ===
	# if print_editors:
	with open(str(config_db_obj["editors"]), 'rb') as editors_pkl_handle:
		editors_dict = pickle.load(editors_pkl_handle)
		print(f"Available editors:\n {list(editors_dict['clinical'].keys())}")
	# if print_base_editors:
	with open(str(config_db_obj["base_editors"]), 'rb') as be_pkl_handle:
		base_editors_dict = pickle.load(be_pkl_handle)
		print(f"Available base editors:\n {list(base_editors_dict['all'].keys())}")



# == Native Modules
# == Installed Modules
# == Project Modules
from build_casoff_input import (check_bulge, parse_bulge)


def parse_nobulge(tmp_processing_filename):
	nobulge_dict = {}
	with open(tmp_processing_filename, 'r') as inf:
		for line in inf:
			entry = line.strip().split(' ')
			if len(entry) > 2 and len(entry[-1]) > 3:
				seq, mm, gid = entry
				nobulge_dict[seq] = [gid, mm]
	return nobulge_dict


def write_casoff_output(raw_casoff_out, casoff_support_file, formatted_output, bulge_check):
	with open(raw_casoff_out) as fi, open(formatted_output, 'w') as fo:
		fo.write('Coordinates\tDirection\tGuide_ID\tBulge type\tcrRNA\tDNA\tMismatches\tBulge Size\n') \

		ot_coords = []
		for line in fi:
			entries = line.strip().split('\t')
			ncnt = 0

			if not bulge_check:
				nobulge_dict = parse_nobulge(casoff_support_file)
				gid, mm = nobulge_dict[entries[0]]
				coord = f'{entries[1]}:{entries[2]}-{int(entries[2]) + len(entries[0])}'
				fo.write(f'{coord}\t{entries[4]}\t{gid}\tX\t{entries[0]}\t{entries[3]}\t{entries[5]}\t0\n')
				ot_coords.append(coord)
			if bulge_check:
				(isreversed,
				 chrom_path,
				 seq_pam,
				 rnabulge_dic,
				 id_dict,
				 len_pam,
				 pattern,
				 bulge_dna,
				 bg_tgts) = parse_bulge(casoff_support_file)

				if isreversed:
					for c in entries[0][::-1]:
						if c == 'N':
							ncnt += 1
							break
					if ncnt == 0:
						ncnt = -len(entries[0])
				else:
					for c in entries[0]:
						if c == 'N':
							ncnt += 1
						else:
							break

				if entries[0] in rnabulge_dic:
					gid = id_dict[entries[0]]
					for pos, query_mismatch, seq in rnabulge_dic[entries[0]]:
						if isreversed:
							tgt = (seq_pam + entries[0][len_pam:len_pam + pos] + seq + entries[0][len_pam + pos:-ncnt],
								   entries[3][:len_pam + pos] + '-' * len(seq) + entries[3][len_pam + pos:-ncnt])
						else:
							tgt = (entries[0][ncnt:ncnt + pos] + seq + entries[0][ncnt + pos:-len_pam] + seq_pam,
								   entries[3][ncnt:ncnt + pos] + '-' * len(seq) + entries[3][ncnt + pos:])
						if query_mismatch >= int(entries[5]):
							start = int(entries[2]) + (ncnt if (not isreversed and entries[4] == "+") or (
									isreversed and ncnt > 0 and entries[4] == "-") else 0)
							coord = f'{entries[1]}:{start}-{int(start) + len(tgt[1])}'
							ot_coords.append(coord)
							fo.write(
								f'{coord}\t{entries[4]}\t{gid}\tRNA\t{tgt[0]}\t{tgt[1]}\t{int(entries[5])}\t{len(seq)}\n')

				else:
					gid = id_dict[entries[0]]
					nbulge = 0
					if isreversed:
						for c in entries[0][:-ncnt][len_pam:]:
							if c == 'N':
								nbulge += 1
							elif nbulge != 0:
								break
						tgt = (seq_pam + entries[0][:-ncnt][len_pam:].replace('N', '-'), entries[3][:-ncnt])
					else:
						for c in entries[0][ncnt:][:-len_pam]:
							if c == 'N':
								nbulge += 1
							elif nbulge != 0:
								break
						tgt = (entries[0][ncnt:][:-len_pam].replace('N', '-') + seq_pam, entries[3][ncnt:])
					start = int(entries[2]) + (ncnt if (not isreversed and entries[4] == "+") or (
							isreversed and ncnt > 0 and entries[4] == "-") else 0)
					btype = 'X' if nbulge == 0 else 'DNA'
					coord = f'{entries[1]}:{start}-{int(start) + len(tgt[1])}'
					ot_coords.append(coord)
					fo.write(
						f'{entries[1]}:{start}-{start + len(tgt[1])}\t{entries[4]}\t{gid}\t{btype}\t{tgt[0]}\t{tgt[1]}\t{int(entries[5])}\t{nbulge}\n')

		editor = gid.split('_')[0]
		print(f'{len(ot_coords)} off targets found for {editor}')


def main():
	# SNAKEMAKE IMPORTS
	# === Inputs ===
	raw_casoff_output_path = str(snakemake.input.casoff_out)
	casoff_support_path = str(snakemake.input.casoff_support)
	# === Outputs ===
	formatted_casoff_out = str(snakemake.output.formatted_casoff_out)
	# === Params ===
	rna_bulge = str(snakemake.params.rna_bulge)
	dna_bulge = str(snakemake.params.dna_bulge)
	maximum_mismatches = str(snakemake.params.max_mismatch)
	PU = str(snakemake.params.casoff_accelerator)

	# Check bulge based on pre-defined Cas-Offinder params
	casoff_params = (maximum_mismatches, rna_bulge, dna_bulge, PU)
	bulge_check = check_bulge(casoff_params)

	write_casoff_output(raw_casoff_output_path, casoff_support_path, formatted_casoff_out, bulge_check)


if __name__ == "__main__":
	main()
